;;------------------------------------------------------------------------------
;; This file is part of Farming Spirits.
;;
;; Farming Spirits is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published 
;; by the Free Software Foundation, either version 3 of the License, 
;; or (at your option) any later version.
;;
;; Farming Spirits is distributed in the hope that it will be useful, but WITHOUT 
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
;; FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
;; for more details.
;;
;; You should have received a copy of the GNU General Public License 
;; along with Farming Spirits. If not, see https://www.gnu.org/licenses/. 
;;------------------------------------------------------------------------------

.module game

.include "../system/update.h.s"
.include "../system/animations.h.s"
.include "../system/render.h.s"
.include "../system/collision.h.s"
.include "../system/lanes.h.s"
.include "../system/hud.h.s"
.include "../constants.h.s"
.include "../macros.h.s"
.include "./menu.h.s"
.include "./entity.h.s"
.include "./entity_manager.h.s"
.include "./entity_templates.h.s"
.include "./interruptions.h.s"
.include "./animations.h.s"
.include "../music/music.h.s"
.include "../system/level.h.s"
.include "../screens/screen_title_z.h.s"
.include "../system/screens.h.s"
.include "../system/sounds.h.s"
.include "./screen_manager.h.s"
.include "../render_cpctelera32x16.h.s"
.include "../sprites/sprites.h.s"

.include "cpctelera.h.s"
.include "../cpctelera_includes.h.s"

.area _DATA

player_attacks_templates:
.dw player_red_bullet_tmpl
.dw player_cyan_bullet_tmpl
.dw player_yellow_bullet_tmpl

player_ptr:
.dw 0x0000
player_health::
.db PLAYER_MAX_HEALTH
player_changed_health:
.db 0
current_player_health_changed_frame_count:
.db PLAYER_DAMAGE_FRAME_COUNT_INIT
current_attack_type_index: ;; (0, 1, 2)
.db 0

_SPR_TUTORIAL_X = (SCREEN_WIDTH_IN_BYTES/2)-(_SPR_TUTORIAL_W/2)
_SPR_TUTORIAL_Y = 8
tutorial:
   .dw #_spr_tutorial_shoot

.area _CODE

.globl cpct_waitHalts_asm



game_call_hl::
   jp (hl)

;/////////////////////////////////////
;// Input:
;// Destroys: AF, HL, BC, DE 
;/////////////////////////////////////
game_set_next_attack_type::
   ;; Increase the attack type index
   ld a, (#level_generator_num_attacks)
   ld b, a
   ld a, (#current_attack_type_index)
   inc a

   ;; if the attack index is equal to the maximum set it to 0
   cp b
   jr nz, game_set_next_attack_type_end
      ld a, #0
   game_set_next_attack_type_end:
      ld (#current_attack_type_index), a

      call sys_hud_set_active_attack
ret

;/////////////////////////////////////
;// Input:
;// Destroys: AF, HL, BC, DE 
;/////////////////////////////////////
game_set_previous_attack_type::
   ;; Decrease the attack type index
   ld a, (#level_generator_num_attacks)
   ld b, a
   ld a, (#current_attack_type_index)
   dec a

   ;; if the attack index is negative set it to the maximun - 1
   jp p, game_set_previous_attack_type_end
      dec b
      ld a, b
   game_set_previous_attack_type_end:
      ld (#current_attack_type_index), a

      call sys_hud_set_active_attack
ret

;/////////////////////////////////////
;// Destroys: AF, HL, BC, DE 
;/////////////////////////////////////
game_add_player_health::
   ;;Set the border green
   cpctm_setBorder_asm #HW_BRIGHT_GREEN
   ld a, #1
   ld (#player_changed_health), a

   ld hl, #player_health
   ld a,  (hl)

   ;; Compare the value to see if the player already has the maximun health
   cp #PLAYER_MAX_HEALTH
   ret z
   inc a
   ld (hl), a

   push ix
   push iy
   call sys_hud_add_heart
   pop iy
   pop ix
ret

;/////////////////////////////////////
;// Destroys: AF, HL, BC, DE 
;/////////////////////////////////////
game_remove_player_health::
   push ix
   push iy
   call sys_sounds_play_hit
   pop iy
   pop ix
   
   ;; Set the border red
   cpctm_setBorder_asm #HW_BRIGHT_RED
   ld a, #1
   ld (#player_changed_health), a

   ld hl, #player_health
   ld a, (hl)

   ;; Compare the value to see if the player already has the minimun health (Just in case it receives two hits in a single frame)
   or a
   ret z
   dec a
   ld (hl), a

   push ix
   push iy
   call sys_hud_remove_heart
   pop iy
   pop ix
ret

;/////////////////////////////////////
;// Input:
;//     IY - Player
;// Destroys: A, HL, BC, DE, IX
;/////////////////////////////////////
game_create_player_bullet::
   call entity_man_create_entity

   ;; Calculate the index in the template attack array
   ld  a, (#current_attack_type_index)
   sla a                                ;; current_attack_index *= 2
   ld b, #0
   ld c, a

   ;; Set hl to point to the current attack template
   ld hl, #player_attacks_templates
   add hl, bc
   ld a, (hl)
   inc hl
   ld h, (hl)
   ld l, a

   ;; Copy the entity data
   ld bc, #ENTITY_SIZE
   ldir

   ;; Set the y position of the bullet to be the y position of the player
   ld a, ENTITY_Y(iy)
   ld ENTITY_Y(ix), a
ret

;/////////////////////////////////////
;// Input:
;//     IY - Enemy
;//     HL - Bullet template
;// Destroys: A, HL, BC, DE, IX
;/////////////////////////////////////
game_create_enemy_bullet::
   ;; Create the entity 
   push hl
   call entity_man_create_entity
   pop  hl
   ldir

   ;; Set the y position of the bullet
   ld a, ENTITY_Y(iy)
   add a, #4            ;; OFFSET
   ld ENTITY_Y(ix), a

   ;; Set the x position of the bullet
   ld a, ENTITY_X(iy)
   add a, ENTITY_W(iy)
   ld ENTITY_X(ix), a
ret

;/////////////////////////////////////
;// Input:
;//     A  - Indice linea
;//     HL - Ptr al template
;// Destroys: A, HL, BC, DE, IX, IY
;/////////////////////////////////////
game_create_enemy::
   ;; Create the entity and copy the template data
   push hl
   call entity_man_create_entity
   pop hl
   ldir 

   ;; Load the line from the ptr array into iy
   sla a ;; a * 2
   ld hl, #lanes_ptr_array
   ld b, #0
   ld c, a
   add hl, bc

   ld c, (hl)
   inc hl
   ld b, (hl)

   ld iy, #0
   add iy, bc

   ;; Set the position of the entity in the lane
   ld a, LANE_POS_X(iy)
   ld ENTITY_X(ix), a
   ld a, LANE_POS_Y(iy)

   ;; Quito 4 en y para poner la vida
   sub #4
   ld ENTITY_Y(ix), a

   ;;Store the entity reference into the lane array
   push ix
   pop bc

   ld LANE_SLOTS_START_SLOT(iy),     c
   ld LANE_SLOTS_START_SLOT + 1(iy), b
ret

;/////////////////////////////////////
;// Input:
;//     IX  - Enemigo
;// Destroys: A, HL, BC, DE
;/////////////////////////////////////
game_kill_enemy::
   ;; Ignore if the enemy is already dead (The lane has killed it)
   ld a, ENTITY_STATE(ix)
   and #ENTITY_STATE_DYING
   ret nz

   ld a, ENTITY_CMPS(ix)
   and #COMPONENT_DEAD_PTR
   jr z, game_kill_enemy_no_dead_ptr_component
      ld l, ENTITY_DEAD_PTR_L(ix)
      ld h, ENTITY_DEAD_PTR_H(ix)
      call game_call_hl
   game_kill_enemy_no_dead_ptr_component:

      ld b, #ANIM_ENEMY_DEATH
      ld c, #ANIM_ENEMY_DEATH_FRAMES
      call game_mark_enemy_as_dead

      ;; Remove the entity from the lane
      ld c, ENTITY_LANE_Y(ix) ;; Line index
      ld e, ENTITY_LANE_X(ix) ;; Entity index
      call sys_lanes_get_lane_entity_slot

      ;;Set the slot to null
      ld (hl), #0
      inc hl
      ld (hl), #0
ret

;/////////////////////////////////////
;// Input:
;//     IX  - Enemigo
;//     B - Estado animacion
;//     C - FRAMES ANIMACION
;// Destroys: A, HL, BC, DE
;/////////////////////////////////////
game_mark_enemy_as_dead::
   ;; Set the entity as dying -> Remove all the components except render, update and animated, and set the new update function
   ld ENTITY_CMPS(ix), #COMPONENT_RENDER | #COMPONENT_UPDATE | #COMPONENT_ANIMATED
   ld ENTITY_ANIM_STATUS(ix), b
   ld ENTITY_UPDATE_COUNTER(ix), c

   ld a, ENTITY_STATE(ix)
   or #ENTITY_STATE_DYING
   ld ENTITY_STATE(ix), a


   ld ENTITY_UPDATE_BEHAVIOUR_L(ix), #<sys_update_behaviour_enemy_dying
   ld ENTITY_UPDATE_BEHAVIOUR_H(ix), #>sys_update_behaviour_enemy_dying
ret

;/////////////////////////////////////
;// Input:
;//     IY  - Entidad
;//      A  - Dead Frames
;/////////////////////////////////////
game_kill_bullet_ix::
   ;; Set the entity as dying -> Remove all the components except render, update and animated, and set the new update function
   ld ENTITY_CMPS(ix), #COMPONENT_RENDER | #COMPONENT_UPDATE | #COMPONENT_ANIMATED
   ld ENTITY_ANIM_STATUS(ix), #ANIM_ENTITY_DEATH
   ld ENTITY_UPDATE_COUNTER(ix), a

   ld ENTITY_UPDATE_BEHAVIOUR_L(ix), #<sys_update_behaviour_entity_dying
   ld ENTITY_UPDATE_BEHAVIOUR_H(ix), #>sys_update_behaviour_entity_dying
ret

;/////////////////////////////////////
;// Input:
;//     IY  - Entidad
;//      A  - Dead Frames
;/////////////////////////////////////
game_kill_bullet_iy::
   ;; Set the entity as dying -> Remove all the components except render, update and animated, and set the new update function
   ld ENTITY_CMPS(iy), #COMPONENT_RENDER | #COMPONENT_UPDATE | #COMPONENT_ANIMATED
   ld ENTITY_ANIM_STATUS(iy), #ANIM_ENTITY_DEATH
   ld ENTITY_UPDATE_COUNTER(iy), a

   ld ENTITY_UPDATE_BEHAVIOUR_L(iy), #<sys_update_behaviour_entity_dying
   ld ENTITY_UPDATE_BEHAVIOUR_H(iy), #>sys_update_behaviour_entity_dying
ret

game_handle_player_damage_taken:
   ld a, (#player_changed_health)
   or a
   ret z

   ld a, (#current_player_health_changed_frame_count)
   dec a
   ld (#current_player_health_changed_frame_count), a
   ret nz

   cpctm_setBorder_asm #HW_BLACK

   ld a, #PLAYER_DAMAGE_FRAME_COUNT_INIT
   ld (#current_player_health_changed_frame_count), a

   xor a
   ld (#player_changed_health), a
ret

game_reset_player_variables::
   xor a
   ld (#player_changed_health), a
   ld (#current_attack_type_index), a

   ld a, #PLAYER_MAX_HEALTH
   ld (#player_health), a

   ld a, #PLAYER_DAMAGE_FRAME_COUNT_INIT
   ld (#current_player_health_changed_frame_count), a
ret

game_draw_tutorial:
   ld hl,   (tutorial)
   ld de, #0
   sbc hl, de
   ret z

   ld a, (level_generator_current_num_level_right)
   cp #1
   jr z, game_draw_tutorial_to_shoot
   cp #2
   jr z, game_draw_tutorial_to_change_spell
   cp #3
   jr z, game_draw_tutorial_to_move
   ld hl, #0x0000
   ld (tutorial), hl
   ret


   game_draw_tutorial_to_shoot:
   ld hl, #_spr_tutorial_shoot
   jr game_draw_tutorial_end
   game_draw_tutorial_to_move:
   ld hl, #_spr_tutorial_move 
   jr game_draw_tutorial_end
   game_draw_tutorial_to_change_spell:
   ld hl, #_spr_tutorial_spell

   game_draw_tutorial_end:
   ld (tutorial), hl

   ld d,   #0xC0
   game_draw_tutorial_loop:
   ld e,   #0
   ld c,    #_SPR_TUTORIAL_X
   ld b,    #_SPR_TUTORIAL_Y
   push de
   call getScreenPtr_32x16
   ex de,   hl
   ld hl,   (tutorial)
   ld c,    #_SPR_TUTORIAL_W
   ld b,    #_SPR_TUTORIAL_H
   call     drawSprite_32x16
   pop de
   ld a, d
   cp #0xC4
   jr z, game_draw_tutorial_end_loop
   ld d, #0xC4
   jr game_draw_tutorial_loop

   game_draw_tutorial_end_loop:
   
   
   ret
   
;;

game_init::
   call sys_lanes_update_lanes_positions
   call sys_lanes_reset
   call game_reset_player_variables

   call entity_man_init
   call sys_render_reset_double_buffer

   ld de, #_song_game
   call cpct_akp_musicInit_asm 

   call sys_hud_reset
   call sys_screens_draw_ingame
   call sys_lanes_draw
   call sys_draw_hud

   call game_draw_tutorial


   ;;Create the player and set the y position
   call entity_man_create_entity
   ld hl, #player_tmpl
   ldir 
   ld (#player_ptr), ix
   call sys_lanes_set_player_position

ret

game_handle_player_death:
   call sys_sounds_play_dead

   ;; Remove the update and collider component and set the dead animation
   ld ix, (#player_ptr)
   ld ENTITY_CMPS(ix), #(COMPONENT_ANIMATED | COMPONENT_RENDER)
   ld ENTITY_ANIM_STATUS(ix), #ANIM_PLAYER_DEATH
   ld a, #ANIM_PLAYER_DEATH_FRAMES - 1

   game_handle_player_death_loop:
      push af

      call sys_animations_update

      call sys_update_behaviour_update
      call sys_render_update
      call entity_man_update

      call sys_render_swap_buffers
      call interruptions_wait_sync

      ;; Exit the loop when the player animation frames have finished
      pop af
      dec a
   jr nz, game_handle_player_death_loop

   ld b, #32
   game_handle_player_death_wait_loop:
      call interruptions_wait_sync
      dec b
   jr nz, game_handle_player_death_wait_loop
   cpctm_setBorder_asm #HW_BLACK
ret

game_play::
   ld hl, #interruptions_game_int_handler
   call interruptions_set_interruption_handler

game_play_loop:
   call sys_render_update
   call entity_man_update
   call level_update
   call sys_animations_update
   call sys_lanes_update
   call sys_collisions_update
   call sys_update_behaviour_update
   
   call game_handle_player_damage_taken
   
   ;; Check if the player has died
   ld a, (#player_health)
   or a
   jr z, game_play_game_over

   ;; Check if the level has been completed
   call level_level_has_been_completed
   jr z, game_play_level_completed

   call sys_render_swap_buffers
   call interruptions_wait_sync

   jr game_play_loop

game_play_level_completed:
   call level_has_next_level
   ;; If there is no next level go to the victory screen
   call z, screen_man_to_victory
   ;; Else -> Go to the next level screen
   call screen_man_to_next_level_screen
game_play_game_over:
   call game_handle_player_death
   call screen_man_to_game_over
