;;------------------------------------------------------------------------------
;; This file is part of Farming Spirits.
;;
;; Farming Spirits is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published 
;; by the Free Software Foundation, either version 3 of the License, 
;; or (at your option) any later version.
;;
;; Farming Spirits is distributed in the hope that it will be useful, but WITHOUT 
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
;; FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License 
;; for more details.
;;
;; You should have received a copy of the GNU General Public License 
;; along with Farming Spirits. If not, see https://www.gnu.org/licenses/. 
;;------------------------------------------------------------------------------

.include "cpctelera.h.s"

;; STRINGS
;;
;; Compilation control directives
;;
;; If true, interrupts can be disabled to get function parameters faster
.equ let_disable_interrupts_for_function_parameters, 0  

;;
;; Constant values
;;
.equ char0_ROM_address, 0x3800   ;; Address where definition of character 0 starts in ROM
.equ GA_port_byte,      0x7F     ;; 8-bit Port of the Gate Array

;;
;; External values
;;
.globl _cpct_mode_rom_status       ;; defined in cpc_mode_rom_status.s
.globl dc_mode0_ct   ;; Colour conversion table (16 values to convert PEN values [0-15] to their screen pixel format)



cpct_drawCharM0_32x16::

   ;; Enable Lower ROM during char copy operation, with interrupts disabled 
   ;; to prevent firmware messing things up
   ;ld     a,(_cpct_mode_rom_status)  ;; [4] A = mode_rom_status (present value)
   ;and    #0b11111011                ;; [2] bit 3 of A = 0 --> Lower ROM enabled (0 means enabled)
   ;ld     b, #GA_port_byte           ;; [2] B = Gate Array Port (0x7F)
   ;di                                ;; [1] Disable interrupts to prevent firmware from taking control while Lower ROM is enabled
   ;out   (c), a                      ;; [3] GA Command: Set Video Mode and ROM status (100)

   ;; Draw the character
   ld     a, e                       ;; [1] A = ASCII Value of the character
   call   cpct_drawCharM0_inner_32x16  ;; [828/837] Does the actual drawing to screen

endDraw:
   ;; After finishing character printing, restore previous ROM and Interrupts status
   ;ld     a, (_cpct_mode_rom_status) ;; [4] A = mode_rom_status (present saved value)
   ;ld     b, #GA_port_byte           ;; [2] B = Gate Array Port (0x7F)
   ;out   (c), a                      ;; [3] GA Command: Set Video Mode and ROM status (100)
   ;ei                                ;; [1] Enable interrupts

ret

cpct_drawCharM0_inner_32x16::
   ;; Calculate the memory address where the 8 bytes defining the character appearance 
   ;; ... start (IX = 0x3800 + 8*ASCII value). char0_ROM_address = 0x3800
   sub #32
   rlca           ;; [1] | A = E = 8*ASCII. 3 RLCA leave A with this
   rlca           ;; [1] |   |hgfedcba| => 3*RLCA => |edcba|hgf|         IXH         IXL
   rlca           ;; [1] | Then we need to move it to IX like this => |00111hgf| |edcba000|
   ld    e, a     ;; [1] \ That will be the final memory address where the definition starts
   and   #0x07    ;; [2] Isolate latest 3 bits of a |00000hgf|
   ;or    #0x38    ;; [2] Add the 3 ones in front, so that the address starts at 0x38xx => |00111hgf|
   ld__ixh_a      ;; [2] Save it to IXH = |00111hgf|
   ld    a, e     ;; [1] Restore A status after 3*RLCA => |edcba|hgf|
   and   #0xF8    ;; [2] Isolate first 5 bits => |edcba|000|
   ld__ixl_a      ;; [2] and save it to IXL = |edcba|000|
   ;; Now IX = |edcba|000||00111hgf| = 0x3800 + 8*ASCII

   ld    bc, #0x40
   add   ix, bc

   ld    bc, #dc_2pxtableM0    ;; [3] BC points to the 2 1-bit pixels to 2 4-bit pixels conversion table

   ;; Draw next line from the character to the screen
nextline:
   ex    de, hl      ;; [1] Put Destination pointer into DE (it is in HL)
   ld     a, (ix)    ;; [5] A = Next Character pixel line definition 
                     ;; .... (8 bits defining 0 = background colour, 1 = foreground)
   ;; Copy the 4 bytes that compose the complete pixel line
   ;; repeating the code for each pair of pixels to maximize speed
.rept 4
   ;; Convert next 2-bits into 1 byte with 2 pixels in screen pixel format
   ;; and copy it to (DE) which is next screen location
   ld    hl, #0      ;; [3] HL = 0
   rlca              ;; [1] /    Put the 2 leftmost bits of A into the two 
   rl    l           ;; [2] | ...rightmost bits of L. This is the combination for the
   rlca              ;; [1] | ...next 2 pixels (BG-BG, BG-FG, FG-BG, FG-FG). We use it
   rl    l           ;; [2] \ ...as index for the dc_2pxtableM0 which gets the actual pixel values.
   add   hl, bc      ;; [3] HL = BC + L (2pxtableM0 + Index = HL Points to the converted pixel values)
   ldi               ;; [5] Copy 2 pixels to the screen, incrementing DE at the same time
   inc   bc          ;; [2] BC is decremented by LDI but we want it to keep pointing to the table, so we add 1 again
.endm

endpixelline:
   ;; Move to next pixel-line definition of the character
   inc__ixl          ;; [2] Next pixel Line (characters are 8-byte-aligned in memory, 
                     ;; ... so we only need to increment IXL, as IXH will not change)
   ld__a_ixl         ;; [2] If next pixel line corresponds to a new character 
                     ;; .... (this is, we have finished drawing our character), ....
   and   #0x07       ;; [2] ... then L % 8 == 0, as it is 8-byte-aligned. 
   ret   z           ;; [2/4] If L % 8 == 0, we have finished drawing the character, else, proceed to next line

   ;; Prepare to copy next line 
   ;;  -- Move DE pointer to the next pixel line on the video memory
   ;; (We save new calculations on HL, because it will be exchanged with DE at the start of nextline: loop)
   ld    hl, #0x800-4      ;; [3] | Next pixel line is 0x800 bytes away in standard video modes
   add   hl, de            ;; [3] | ..but DE has already being incremented 4 times. So add 0x800-4 to
                           ;;       ..DE to make it point to the start of the next pixel line in video memory
   ;; Check if new address has crossed character boundaries (every 8 pixel lines)
   ld     a, h             ;; [1] A = H (top 8 bits of video memory address)
   and   #0x38             ;; [2] We check if we have crossed memory boundary (every 8 pixel lines)
   jr    nz, nextline      ;; [2/3]  by checking the 4 bits that identify present memory line. 
                           ;; .... If 0, we have crossed boundaries
boundary_crossed:
   ld    de, #0xC040       ;; [3] | HL = HL + 0xC050: Relocate DE pointer to the start of the next pixel line in video memory
   add   hl, de            ;; [3] \ (Remember that HL and DE will be exchanged at the start of nextline:)
   jr    nextline          ;; [3] Jump to continue with next pixel line

;; Conversion table from 2 1-bit pixels to mode 0 2 4-bit pixels. Essentially, there are 4
;; possible combinations with 2 pixels and 2 colours: (00, 01, 10, 11 == BG-BG, BG-FG, FG-BG, FG-FG)
;; We reserve here 4 bytes that will be filled in by <cpct_setDrawCharM0>
;;
dc_2pxtableM0:: .db 0x00, 0x40, 0x80, 0xC0   ;; Default colours BG=0, FG=1


cpct_setDrawCharM0_32x16::
   ;; Get Colour values converted to Screen Pixel Format
   ;;
   ld    de, #dc_mode0_ct  ;; [3]  DE = Pointer to the start of the Conversion Colour Table
  
   ;; Convert foreground colour and store it into B
   ld     b, h             ;; [1] Save Background Pen into C for later use
   ld     h, #0            ;; [2] | HL = L    (Index of the colour into the table [0-15])
   add   hl, de            ;; [3] HL = DE + L ([Table base + Index], Points to the converted value)
   ld     c, (hl)          ;; [2] C = Conversion of foreground colour to screen pixel format
  
   ;; Convert background colour and store it into C
   ld     h, #0            ;; [2] |
   ld     l, b             ;; [1] | HL = B    (Index of the colour into the table [0-15])
   add   hl, de            ;; [3] HL = DE + B ([Table base + Index], Points to the converted value)
   ld     b, (hl)          ;; [2] B = Conversion of background colour to screen pixel format

   ;; Produce the 4 combinations for pairs of pixels (BG-BG, BG-FG, FG-BG, FG-FG)
   ;; and store them in the placeholders inside <cpct_drawCharM0>
   ld    hl, #dc_2pxtableM0;; [3] HL Points to the 4-byte colour conversion table inside cpct_drawCharM0

   ;; Calculate and store BG-BG mode 0 pixel pair
   ld     a, b             ;; [1] A = Right Pixel is BG coloured
   rlca                    ;; [1] Shift left so A = Left pixel is BG coloured
   or     b                ;; [1] A = Left and Right pixels are BG coloured
   ld    (hl), a           ;; [2] Store BG-BG pair of mode 0 pixels in the table
   
   ;; Calculate and store BG-FG mode 0 pixel pair
   inc   hl                ;; [2] Point to next pair (BG-FG) in the table
   xor    b                ;; [1] Set Right Pixel to 0 (Its value is equal to B)
   or     c                ;; [1] Left Pixel: BG, Right Pixel: FG
   ld    (hl), a           ;; [2] Store BG-FG pair of mode 0 pixels in the table

   ;; Calculate and store FG-BG mode 0 pixel pair
   inc   hl                ;; [2] Point to next pair (FG-BG) in the table
   rlca                    ;; [1] Left Pixel: FG, Right Pixel: xx
   and   #0xAA             ;; [2] Right Pixel: 00
   or    b                 ;; [1] Left Pixel: FG, Right Pixel: BG
   ld    (hl), a           ;; [2] Store FG-BG pair of mode 0 pixels in the table

   ;; Calculate and store FG-FG mode 0 pixel pair
   inc   hl                ;; [2] Point to next pair (FG-BG) in the table
   xor    b                ;; [1] Left Pixel: FG, Right Pixel: 0
   or     c                ;; [1] Left Pixel: FG, Right Pixel: FG
   ld    (hl), a           ;; [2] Store FG-FG pair of mode 0 pixels in the table

   ;; Everything set up. Now return
   ret                     ;; [3] Return



;/////////////////////////////////////
;// FUNCTION
;// Sustitucion de la de cpctelera para pantalla de 32x16
;// Input:
;//     DE - memory     	Video memory pointer to the upper left box corner byte
;//     A  - colour_pattern 1-byte colour pattern (in screen pixel format) to fill the box with
;//     C  - width  	    Box width in bytes [1-64] (Beware!  not in pixels!)
;//     B  - height     	Box height in bytes (>0)
;/////////////////////////////////////
drawSolidBox_32x16::
   ld     l, a             ;; [1] L=A (Colour pattern)
   ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
   ;; COPY2HL_n_INC Macro 
   ;;  Copy Register To (HL) and Increment HL
   ;; INPUT:
   ;;    reg8 - any 8-bits register
   ;;
   .macro COPY2HL_n_INC reg8
   ld   (hl), reg8         ;; [2] (HL) = reg8
   inc   hl                ;; [2] HL++  
   .endm

   ;; DrawSolidBox code starts here (immediately after getting parameters from stack)
   
   ;; Modify code using width to jump to ensure
   ;; that only width bytes are copied each line 
   ld     a, #128          ;; [2] We need to jump 128 bytes (64 COPY2HL_n_INC*2 bytes) minus the width of the sprite * 2 (2B)
   sub    c                ;; [1]  to do as much COPY2HL_n_INC as bytes the Sprite is wide
   sub    c                ;; [1]
   ld (_jr_offset), a      ;; [4] Modify JR data to create the jump we need
   
   ld     c, l             ;; [1] C=L (Colour pattern)
   ld     h, d             ;; [1] / HL = DE
   ld     l, e             ;; [1] \ HL Points to Video Memory, and DE holds a copy which won't be modified

   _next_line:
   _jr_offset = .+1
   jr__0                   ;; [3] Self modifying instruction: the '00' will be substituted by the required jump forward. 
                           ;; ... (Note: Writting JR 0 compiles but later it gives odd linking errors)  
   ;; 64 COPY2HL_n_INC c, which are able to copy up to 64 bytes each time.
   ;; That means that each Sprite line should be 64 bytes width at most.
   ;; The JR instruction at the start makes us ignore the COPY2HL_n_INC we don't need 
   ;; (jumping over them) That ensures we will be doing only as much COPY2HL_n_INC 
   ;; as bytes our sprite is wide.
   .rept 64
   COPY2HL_n_INC c         ;; [4] (HL) = c, ++HL { Repeated 64 times }
   .endm
   
   dec    b                ;; [1] Another line finished: we discount it from B
   ret    z                ;; [2/4] If that was the last line, we safely return

   ;; Jump destination pointer to the start of the next line in video memory
   ld     a, #0x8          ;; [2] / HL = DE = DE + 0x800
   add    d                ;; [1] | DE holds the pointer to the start of this line
   ld     h, a             ;; [1] | Adding 0x800 makes HL point to the start of
   ld     d, a             ;; [1] | the next line to be filled in with values
   ld     l, e             ;; [1] \ DE holds a copy which won't be modified
   ;; We check if we have crossed video memory boundaries (which will happen every 8 lines). 
   ;; ... If that happens, bits 13,12 and 11 of destination pointer will be 0
   and   #0x38             ;; [2] leave out only bits 13,12 and 11 from new memory address (00xxx000 00000000)
   jp    nz, _next_line    ;; [3] If any bit from {13,12,11} is not 0, we are still inside 
                           ;; ... video memory boundaries, so proceed with next line

   ;; Every 8 lines, we cross the 16K video memory boundaries and have to
   ;; reposition destination pointer. That means our next line is 16K-0x50 bytes back
   ;; which is the same as advancing 48K+0x50 = 0xC050 bytes, as memory is 64K 
   ;; and our 16bit pointers cycle over it
   ld    hl, #0xC040       ;; [3] We advance destination pointer to next line
   add   hl, de            ;; [3] HL = DE + 0xC050
   ld     d, h             ;; [1] / DE = HL
   ld     e, l             ;; [1] \
   jp   _next_line         ;; [3] Continue copying   
   
   ret
;/////////////////////////////////////

;/////////////////////////////////////
;// FUNCTION
;// Input:
;//     HL - ptr al sprite
;//     DE - ptr a memoria de video
;//     B  - Height
;//     C  - Width
;/////////////////////////////////////
drawSprite_32x16::
   
   ld a, c             ;; cambiamos width       
   ld (spr_width), a
   neg
   ld (spr_width_neg), a
   ld__ixl_b ;; ld ixl, b
   next_line:
   spr_width = .+1
   ;; Copy a sprite row
   ld bc, #0
   ldir            ;; *DE = *HL; HL++, DE++, BC-- (bc = contador)
   dec__ixl        ;; dec ixl (parte baja ix)
   ;;; if (--height == 0 ) return
      ret z
   ;;; }
   ;; Jump to next pixel row
   ex de, hl
   spr_width_neg = . + 1
   ld bc, #0x07FF       ;; 0x0800 - width
   add hl, bc
   ex de, hl
   ;; DE += 07FC - ancho //7FC

   ld a, #0x38
   and d
   jr nz, next_line

   ;; Jump to next pixel row
   ex de, hl
   ld bc, #-8*0x800 + 0x40       ;; 0x800 - width
   add hl, bc
   ex de, hl

   jr next_line

   ret


;/////////////////////////////////////

;/////////////////////////////////////
;// FUNCTION
;//   DIRECCION 0XBC -> CRTC Registro
;//
;// Input:
;//     DE - screen_start	Pointer to the start of the screen (or a backbuffer)
;//     C  - x	[0-79] Byte-aligned column starting from 0 (x coordinate, in bytes)
;//     B  - y	[0-199] row starting from 0 (y coordinate)
;// Return
;//     HL - Pointer to memory addres
;/////////////////////////////////////
getScreenPtr_32x16::

   ;; Before starting, some terminology to understand comments,
   ;;    Screen is composed of 25 Character Rows (R: Row)
   ;;    Each Character Row is composed of 8 pixel lines (L: Line)
   ;;    CPU Registers are named with a small r in-front (rA, rB, rH, rL, rHL...)
   ;;
   ;; 1st of all, split Y-coordinate (rB) into:
   ;;    rH = B % 8         ;; <<- Line [0-7] inside the character Row
   ;;    rL = int(B / 8)    ;; <<- Screen character Row [0-24]
   ;;
   ;; We can see Y-coordinate as including this two numbers (R: Row, L: Line) 
   ;; split into its first 5-bits and its last 3-bits:
   ;;    Y-coordinate == [R|R|R|R|R|L|L|L]
   ;;                bit7-/             \-bit0
   ;;

   ;; Extract line number (L) from Y-coordinate
   ;; And multiply it by 256. rHL = 256 * L
   ld    a, b     ;; [1] rA = Y-Coordinate
   and   #0x07    ;; [2] /
   ld    h, a     ;; [1] \ rH = Y % 8      ;; << rH contains Line number [0-7] inside character Row
                  ;;     Putting Line Number into rH is similar to putting it 
                  ;;     into rHL and then shifting it 8-bits left. Same as 256*rHL.
                  ;;     Therefore, this is like doing HL += 256*L

   ;; Now extract Screen Character Row (R) from Y-Coordinate
   xor   b        ;; [1] / rA = ( rB and #0x07 ) xor rB  =  rB and #0xF8
   ld    l, a     ;; [1] \ rL = 8*int(Y/8) ;; << L contains Screen Character Row multiplied by 8
                                          ;    as bits are shifted 3-bits to the left because
                                          ;;    the 3-least-significant-bits had the line number (L)

   ;; Now rHL = 256*L + 8*R  { rH = Y%8 = L ||| rL = 8*int(Y/8) = 8*R }
   ;; We need to calculate 2048*L + 80*R (0x0800*L + 0x50*R) cambiar 0x50 por 0x40 = 64
   ;; This can be factored into 8*(256*L + 10*R)
   ;; So first, make rL contain 10*R and then we will have rHL = 256*L + 10*R
   ;; And then we will only need to multiply rHL by 8
   
   ;; rA = rL = 8*int(Y/8)
   ;; rA' = rA / 4 = 2 * int(Y/8)
   ;rrca           ;; [1] / rA' = rA / 4 = 2*int(Y/8)
   ;rrca           ;; [1] \ 
   ;;; rA = rL = 8*int(Y/8)
   ;; rL' = rL + rA' = 10 * int(Y/8)
   ;add   a, l     ;; [1] / rL = rL + rA' = 8*int(Y/8) + 2*int(Y/8) = 10*int(Y/8)
   ld    l, a     ;; [1] \ 


   ;;; 64*HL(?)
   ;; Now rHL = 256*L + 10*R
   ;; Just multiply rHL' = 8*rHL = 8*(256*L + 10*R) = 2048*L + 80*R
   add   hl, hl   ;; [3] / rHL' = 8*rHL
   add   hl, hl   ;; [3] | rHL' = 2048*L + 80*R
   add   hl, hl   ;; [3] \ 

   ;; To complete the calculations, we only need to add the X-Coordinate
   ;; and the pointer to the start of the video buffer
   
   ;; Add up X coordinate
   ld     b, #0   ;; [2] / As rC = X-Coordinate, having rB=0 makes rBC = X-Coordinate
   add   hl, bc   ;; [3] \ rHL' = rHL + X

   ;; Add up screen start address we still keep in DE
   add   hl, de   ;; [3] rHL' = rHL + screen_start

   ;; HL now contains the pointer to the byte in the video buffer. Just return it
   ret            ;; [3] return rHL = Pointer to the video buffer at (X,Y) byte coordinates



;/////////////////////////////////////

.include "strings/strings.s"
cpct_drawStringM0_32x16::

      ;; Enable Lower ROM during char copy operation, with interrupts disabled 
      ;; to prevent firmware messing things up
      ;ld     a,(_cpct_mode_rom_status) ;; [4] A = mode_rom_status (present value)
      ;and    #0b11111011               ;; [2] bit 3 of A = 0 --> Lower ROM enabled (0 means enabled)
      ;ld     b, #GA_port_byte          ;; [2] B = Gate Array Port (0x7F)
      ;di                               ;; [1] Disable interrupts to prevent firmware from taking control while Lower ROM is enabled
      ;out   (c), a                     ;; [3] GA Command: Set Video Mode and ROM status (100)

      jr    firstChar                  ;; [3] Jump to first char (Saves 1 jr back every iteration)

   nextChar:
      ;; Draw next character
      push  hl                         ;; [4] Save HL
      call  cpct_drawCharM0_inner_32x16;; [5 + 824/832] Draws the next character
      pop   hl                         ;; [3] Recover HL 

      ;; Increment Pointers
      ld    de, #4                     ;; [3] /
      add   hl, de                     ;; [3] | HL += 4 (point to next position in video memory, 8 pixels to the right)
      inc   iy                         ;; [3] IX += 1 (point to next character in the string)

   firstChar:
      ld     a, (iy)                   ;; [5] A = next character from the string
      or     a                         ;; [1] Check if A = 0
      jr    nz, nextChar               ;; [2/3] if A != 0, A is next character, draw it, else end

   endstring:
      ;; After finishing character drawing, restore previous ROM and Interrupts status
      ;ld     a, (_cpct_mode_rom_status) ;; [4] A = mode_rom_status (present saved value)
      ;ld     b, #GA_port_byte           ;; [2] B = Gate Array Port (0x7F)
      ;out   (c), a                      ;; [3] GA Command: Set Video Mode and ROM status (100)
      ;ei                                ;; [1] Enable interrupts

   ;; IX/IY Restore and Return provided by bindings
ret
;/////////////////////////////////////